/*******************************************************************
 *
 *	File:		EmuZoomDevice.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#define OEMRESOURCE 
#include <windows.h>
#include <e32def.h>
#include <e32std.h>
#include <e32base.h>
#include <gdi.h>

#include "EmuZoomDevice.h"

/***************************************************
 *
 * This emulates the second framebuffer by creating 
 * a WIN32 window and bitblt-ing a bitmap to it.
 *
 ***************************************************/


#define FB_SHADOW_HEAP_SIZE		(64 * 1024)

static class CEmuZoomDevice*	sSelf = NULL;

class CEmuZoomDevice : public CEmuZoomDeviceBase
{
public:
	static CEmuZoomDevice* NewL(const TSize& aSize, TUint8* aPixels, TInt aBpp);
	~CEmuZoomDevice();

	virtual void UpdateFrame();

protected:
	CEmuZoomDevice(const TSize& aSize, TUint8* aPixels, TInt aBpp);
	void ConstructL();
	void CreateThreadL();
	void CreateBitmapL();
	void SetEmuPixelsToBitmap();

	HWND		iWindow;
	RThread		iThread;
	TSize		iSize;
	TUint8*		iPixels;
	TInt		iBpp;

		// run in the context of the child thread
	friend TInt32 __stdcall WndProc(HWND aHwnd, TUint aMessage, TUint aWParam, TInt32 aLParam);
	static TInt TheThreadMain(TAny* aAny);
	void ThreadMain();
	TInt32 HandleWin32Event(HWND aHwnd, TUint aMessage, TUint aWParam, TInt32 aLParam);
	TBool	iMouseDown;

		// shared by the 2 threads
	HBITMAP				iBitmap;
	HDC					iMemDc;
	CRITICAL_SECTION	iCriticalSection;
};


CEmuZoomDeviceBase*	CreateZoomDeviceL(const TSize& aSize, TUint8* aPixels, TInt aBpp)
{
	return CEmuZoomDevice::NewL(aSize, aPixels, aBpp);
}

CEmuZoomDevice*	CEmuZoomDevice::NewL(const TSize& aSize, TUint8* aPixels, TInt aBpp)
{
	CEmuZoomDevice*	device = new(ELeave)CEmuZoomDevice(aSize, aPixels, aBpp);
	CleanupStack::PushL(device);
	device->ConstructL();
	CleanupStack::Pop();
	return device;
}

CEmuZoomDevice::CEmuZoomDevice(const TSize& aSize, TUint8* aPixels, TInt aBpp) :
	iSize(aSize), iPixels(aPixels), iBpp(aBpp)
{
	sSelf = this;
}

CEmuZoomDevice::~CEmuZoomDevice()
{
	iThread.Terminate(KErrNone);
	iThread.Close();
	if (iBitmap)
		::DeleteObject(iBitmap);
	if (iMemDc)
		::DeleteObject(iMemDc);
	::DeleteCriticalSection(&iCriticalSection);
}


void CEmuZoomDevice::ConstructL()
{
	::InitializeCriticalSection(&iCriticalSection);

	CreateBitmapL();
	CreateThreadL();
}

void CEmuZoomDevice::CreateThreadL()
{
	User::LeaveIfError(iThread.Create(_L("FBSHADOW"), CEmuZoomDevice::TheThreadMain, KDefaultStackSize, FB_SHADOW_HEAP_SIZE, FB_SHADOW_HEAP_SIZE, this));
	iThread.Resume();
}


void CEmuZoomDevice::CreateBitmapL()
{
	HDC dc = ::GetDC(NULL);
	iMemDc = ::CreateCompatibleDC(dc);
	iBitmap = ::CreateCompatibleBitmap(dc, iSize.iWidth, iSize.iHeight);
	::SelectObject(iMemDc, iBitmap);
}


TInt32 __stdcall WndProc(HWND aHwnd, TUint aMessage, TUint aWParam, TInt32 aLParam)
{
	TInt32	rval =0;
	if (sSelf)
		rval = sSelf->HandleWin32Event(aHwnd, aMessage, aWParam, aLParam);
	return rval;
}

TInt32 CEmuZoomDevice::HandleWin32Event(HWND aHwnd, TUint aMessage, TUint aWParam, TInt32 aLParam)
{
	switch (aMessage)
	{
		case WM_CREATE:
			return 0;

		case WM_PAINT:
		{
			RECT	clientRect;
			::GetClientRect(iWindow, &clientRect); 

			PAINTSTRUCT ps;
			HDC winDc = ::BeginPaint(iWindow, &ps);
			::EnterCriticalSection(&iCriticalSection);
			::BitBlt(winDc, 0, 0, clientRect.right, clientRect.bottom, iMemDc, 0, 0, SRCCOPY);
			::LeaveCriticalSection(&iCriticalSection);
			::EndPaint(iWindow, &ps);
			return 0;
		}
	
		case WM_LBUTTONDOWN:
		{
			TRawEvent	event;
			event.Set(TRawEvent::EButton1Down, LOWORD(aLParam), HIWORD(aLParam));
			UserSvr::AddEvent(event);
			iMouseDown = ETrue;
			return 0;
		}
		
		case WM_MOUSEMOVE:
		{
			if (iMouseDown)
			{
				TRawEvent	event;
				event.Set(TRawEvent::EPointerMove, LOWORD(aLParam), HIWORD(aLParam));
				UserSvr::AddEvent(event);
			}
			return 0;
		}

		case WM_LBUTTONUP:
		{
			TRawEvent	event;
			event.Set(TRawEvent::EButton1Up, LOWORD(aLParam), HIWORD(aLParam));
			UserSvr::AddEvent(event);
			iMouseDown = EFalse;
			return 0;
		}

		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
	}

	return ::DefWindowProc(aHwnd, aMessage, aWParam, aLParam);
}


TInt CEmuZoomDevice::TheThreadMain(TAny* aAny)
{
	((CEmuZoomDevice*)aAny)->ThreadMain();
	return 0;
}


void CEmuZoomDevice::ThreadMain()
{
	WNDCLASS	wndclass;
	wndclass.style = CS_HREDRAW|CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = NULL;
	wndclass.hIcon = ::LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = ::LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH) ::GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = "EmuZoomClass";

	::RegisterClass(&wndclass);

	DWORD		winStyle = WS_OVERLAPPEDWINDOW;
	RECT		rect;
	rect.left = rect.top = 0;
	rect.right = iSize.iWidth;
	rect.bottom = iSize.iHeight;
	::AdjustWindowRect(&rect, winStyle, FALSE);
	rect.bottom -= rect.top;
	rect.right -= rect.left;
	rect.left = rect.top = 0;
 
	iWindow = ::CreateWindow(
		"EmuZoomClass",
		"EmuZoom",
		winStyle,
		rect.left,
		rect.top,
		rect.right,
		rect.bottom,
		NULL,
		NULL,
		NULL,
		NULL);

	::ShowWindow(iWindow, SW_SHOW);
	::UpdateWindow(iWindow);

	MSG			msg;
	while (::GetMessage(&msg, NULL, 0, 0))
	{
		::TranslateMessage(&msg);
		::DispatchMessage(&msg);
	}
}


void CEmuZoomDevice::SetEmuPixelsToBitmap()
{
	::EnterCriticalSection(&iCriticalSection);

	TUint16*	pixels = (TUint16*) iPixels;
	for (TInt y=0 ; y<iSize.iHeight ; y++)
	{
		for (TInt x=0 ; x<iSize.iWidth ; x++)
		{
			TRgb		emuColor = (iBpp == 12) ? TRgb::Color4K(*pixels++) : TRgb::Color64K(*pixels++);
			COLORREF	winColor = RGB(emuColor.Red(), emuColor.Green(), emuColor.Blue());
			::SetPixel(iMemDc, x, y, winColor);
		}
	}
	::LeaveCriticalSection(&iCriticalSection);
}

void CEmuZoomDevice::UpdateFrame()
{
	SetEmuPixelsToBitmap();

	RECT	clientRect;
	::GetClientRect(iWindow, &clientRect); 
	::InvalidateRect(iWindow, &clientRect, FALSE); 
	::UpdateWindow(iWindow);
}
