/*******************************************************************
 *
 *	File:		CFrameBufferMgr.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32std.h>
#include <e32base.h>
#include <e32hal.h>
#include <GraphicsAccelerator.h>

#include "FrameBufferMgr.h"



/******************************************
 *
 * CEmuFrameBufferMgr
 *
 ******************************************/

#ifndef __MARM__
#include "EmuZoomDevice.h"
class CEmuFrameBufferMgr : public CFrameBufferMgrBase
{
public:
	static CEmuFrameBufferMgr* NewL(CFrameTicker* aFrameTicker);
	~CEmuFrameBufferMgr();

public:		// from CFrameBufferMgrBase
	virtual void RemapFrameBuffer();
	virtual void RestoreFrameBuffer();
	
	virtual TUint8*	FrameBuffer();
	TUint8*	ShadowFrameBuffer();

public:		// implements MFrameTickObserver
	virtual void FrameTick(){UpdateFrame();}

protected:
	CEmuFrameBufferMgr(CFrameTicker* aFrameTicker) : iFrameTicker(aFrameTicker){}
	void ConstructL();
	void UpdateFrame();
	TInt ColorDepth(TDisplayMode aDisplayMode);

	TUint8*			iFb;
	TUint8*			iShadowFb;
	TSize			iSize;
	TDisplayMode	iDisplayMode;
	TInt			iColorDepth;
	CFrameTicker*	iFrameTicker;

	CEmuZoomDeviceBase*	iZoomDevice;
};


CEmuFrameBufferMgr::~CEmuFrameBufferMgr()
{
	if (iShadowFb)
		User::Free(iShadowFb);
}


CEmuFrameBufferMgr* CEmuFrameBufferMgr::NewL(CFrameTicker* aFrameTicker)
{
	CEmuFrameBufferMgr*	frameBufferMgr = new(ELeave)CEmuFrameBufferMgr(aFrameTicker);
	CleanupStack::PushL(frameBufferMgr);
	frameBufferMgr->ConstructL();
	CleanupStack::Pop();
	return frameBufferMgr;
}

void CEmuFrameBufferMgr::ConstructL()
{
	TAcceleratedBitmapInfo	info;
	RHardwareBitmap	hardwareBitmap(-1);
	User::LeaveIfError(hardwareBitmap.GetInfo(info));

	iFb = info.iAddress;
	iSize = info.iSize;
	iDisplayMode = info.iDisplayMode;
	iColorDepth = ColorDepth(iDisplayMode);
	
	iShadowFb = (TUint8*) User::AllocL(iSize.iWidth * iSize.iHeight * 2);
}


void CEmuFrameBufferMgr::RemapFrameBuffer()
{
	TRAPD(err, iZoomDevice = CreateZoomDeviceL(iSize, iShadowFb, iColorDepth));
	iFrameTicker->AddFrameTickObserver(this);
}

void CEmuFrameBufferMgr::RestoreFrameBuffer()
{
	iFrameTicker->RemoveFrameTickObserver(this);
	delete iZoomDevice;
	iZoomDevice = NULL;
}
	
TUint8*	CEmuFrameBufferMgr::FrameBuffer()
{
	return iFb;
}

TUint8*	CEmuFrameBufferMgr::ShadowFrameBuffer()
{
	return iShadowFb;
}

void CEmuFrameBufferMgr::UpdateFrame()
{
	if (iZoomDevice)
		iZoomDevice->UpdateFrame();
}

TInt CEmuFrameBufferMgr::ColorDepth(TDisplayMode aDisplayMode)
{
	const TInt colorDepth[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			24,
			32,
			12
		};
	return colorDepth[aDisplayMode];
}
#endif			/* !__MARM__ */

/*****************************************************
 *
 * CTargetFrameBufferMgr
 *
 *****************************************************/

class CTargetFrameBufferMgr : public CFrameBufferMgrBase
{
public:
	static CTargetFrameBufferMgr* NewL();
	~CTargetFrameBufferMgr();

public:		// from CFrameBufferMgrBase
	void RemapFrameBuffer();
	void RestoreFrameBuffer();
	
	TUint8*	FrameBuffer() {return iFb;}
	TUint8*	ShadowFrameBuffer() {return iShadowFb;}

public:		// implements MFrameTickObserver
	virtual void FrameTick(){}
protected:
	CTargetFrameBufferMgr(){};
	void ConstructL();
	void LoadDriverL();
	void FrameBufferInfo();

	void CreateShadowFrameBufferL();

	void RemapShadowFrameBuffer();
	void RestoreShadowFrameBuffer();

	void RemapLinnearToPhysicalFrameBuffer(TLinAddr aLinAddr);
	void RemapLinnearToChunkPhysPages(TLinAddr aLinAddr);

	enum
	{
		EFlagsDriverOpened		= 0x0001,
		EFlagsChunkCreated		= 0x0002,
		EFlagsShadowFbRemapped	= 0x0004
	};

	TInt			iFlags;

	RZoomDriver		iZoomDriver;

	TInt			iColorDepth;
	TSize			iScreenSize;
	TInt			iFbByteSize;

	TUint8*			iFb;
	TPhysAddr		iPhysFb;

	TUint8*			iShadowFb;
	RChunk			iChunk;
	TPhysAddr*		iChunkPhysPageTable;
	TInt			iNoOfPages;
};


CTargetFrameBufferMgr::~CTargetFrameBufferMgr()
{
	if (iFlags & EFlagsShadowFbRemapped)
		RestoreShadowFrameBuffer();
	
	delete[] iChunkPhysPageTable;
	
	if (iFlags & EFlagsChunkCreated)
		iChunk.Close();

	if (iFlags & EFlagsDriverOpened)
	{
		iZoomDriver.Close();
		User::FreeLogicalDevice(LDD_NAME);
	}
}


CTargetFrameBufferMgr* CTargetFrameBufferMgr::NewL()
{
	CTargetFrameBufferMgr*	frameBufferMgr = new(ELeave)CTargetFrameBufferMgr;
	CleanupStack::PushL(frameBufferMgr);
	frameBufferMgr->ConstructL();
	CleanupStack::Pop();
	return frameBufferMgr;
}


void CTargetFrameBufferMgr::ConstructL()
{
	LoadDriverL();
	FrameBufferInfo();
	CreateShadowFrameBufferL();
	RemapShadowFrameBuffer();
}


void CTargetFrameBufferMgr::LoadDriverL()
{
	TVersion	version(1, 0, 0);
	User::FreeLogicalDevice(LDD_NAME);	// allows for replacing driver without reboot
	User::LoadLogicalDevice(LDD_NAME);	// ignore error, may already be loaded
	User::LeaveIfError(iZoomDriver.Open(version));
	iFlags |= EFlagsDriverOpened;
}


void CTargetFrameBufferMgr::FrameBufferInfo()
{
	TScreenInfoV01			screenInfo;
	TPckg<TScreenInfoV01>	sI(screenInfo);
	UserSvr::ScreenInfo(sI);
	iScreenSize = screenInfo.iScreenSize;
	iFb = (TUint8*) screenInfo.iScreenAddress;
	iColorDepth = 2;		// valid assumption for SEMC P800/P900

	iZoomDriver.LinearToPhysical((TLinAddr)iFb, &iPhysFb);
}

void CTargetFrameBufferMgr::CreateShadowFrameBufferL()
{
	iFbByteSize = iScreenSize.iWidth * iScreenSize.iHeight * iColorDepth;
	iNoOfPages = (iFbByteSize + PAGE_SIZE - 1) >> PAGE_SIZE_SHIFT;
	TInt	size = iNoOfPages * PAGE_SIZE;
	User::LeaveIfError(iChunk.CreateLocal(size, size));
	iShadowFb = iChunk.Base();
	iFlags |= EFlagsChunkCreated;

	iChunkPhysPageTable = new(ELeave)TPhysAddr[iNoOfPages];
	TUint8*		page = iShadowFb;
	for (TInt i=0 ; i<iNoOfPages ; i++)
	{
		iZoomDriver.LinearToPhysical((TLinAddr)page, &iChunkPhysPageTable[i]);
		page += PAGE_SIZE;
	}
}

void CTargetFrameBufferMgr::RestoreShadowFrameBuffer()
{
	RemapLinnearToChunkPhysPages((TLinAddr)iShadowFb);
}

void CTargetFrameBufferMgr::RemapShadowFrameBuffer()
{
	RemapLinnearToPhysicalFrameBuffer((TLinAddr)iShadowFb);
	iFlags |= EFlagsShadowFbRemapped;
}


void CTargetFrameBufferMgr::RemapFrameBuffer()
{
	RemapLinnearToChunkPhysPages((TLinAddr)iFb);
	Mem::Copy(iFb, iShadowFb, iFbByteSize);
}


void CTargetFrameBufferMgr::RestoreFrameBuffer()
{
	Mem::Copy(iShadowFb, iFb, iFbByteSize);
	RemapLinnearToPhysicalFrameBuffer((TLinAddr)iFb);
}

void CTargetFrameBufferMgr::RemapLinnearToChunkPhysPages(TLinAddr aLinAddr)
{
	for (TInt i=0 ; i<iNoOfPages ; i++)
	{
		iZoomDriver.RemapLinearAddr(aLinAddr, iChunkPhysPageTable[i], ETrue, ETrue);
		aLinAddr += PAGE_SIZE;
	}
}


void CTargetFrameBufferMgr::RemapLinnearToPhysicalFrameBuffer(TLinAddr aLinAddr)
{
	TPhysAddr	physFb = iPhysFb;
	for (TInt i=0 ; i<iNoOfPages ; i++)
	{
		iZoomDriver.RemapLinearAddr(aLinAddr, physFb, EFalse, EFalse);
		aLinAddr += PAGE_SIZE;
		physFb += PAGE_SIZE;
	}
}


/******************************************
 *
 * Factory
 *
 ******************************************/

CFrameBufferMgrBase* CreateFrameBufferMgrL(CFrameTicker* aFrameTicker)
{
#ifdef __MARM__
	return CTargetFrameBufferMgr::NewL();
#else
	return CEmuFrameBufferMgr::NewL(aFrameTicker);
#endif
}