/*******************************************************************
 *
 *	File:		InputManager.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32def.h>
#include <e32std.h>
#include <e32base.h>
#include <e32svr.h>

#include "InputManager.h"
#include "ControllerBase.h"
#include "FrameTicker.h"

/**********************************************************
 *
 * class InputManager
 * - Provides keyrepeats when needed
 * - Ensures key-ups get consumed when key-downs have been
 *   consumed
 *
 **********************************************************/

TBool TInputManager::OfferRawEvent(const TRawEvent& aRawEvent)
{
	TBool	consumed = EFalse;
	if (iController)
	{
		if (!iController->ModifyRawEvent((TRawEvent&)aRawEvent))
		{
			switch (aRawEvent.Type())
			{
				case TRawEvent::EKeyUp:
					if (HasConsumedKeyDown(aRawEvent.ScanCode()))
						consumed = ETrue;
					break;

				case TRawEvent::EKeyDown:
				{
					TBool	wantRepeat = EFalse;
					consumed = iController->HandleKeyEvent(aRawEvent.ScanCode(), wantRepeat);
					if (consumed)
						MarkKeyDown(aRawEvent.ScanCode(), wantRepeat);
					break;
				}

				case TRawEvent::EButton1Down:
					consumed = iController->HandlePointerDown(aRawEvent.Pos());
					break;

				case TRawEvent::EButton1Up:
					consumed = iController->HandlePointerUp(aRawEvent.Pos());
					break;

				case TRawEvent::EPointerMove:
					consumed = iController->HandlePointerMove(aRawEvent.Pos());
					break;

				default:
					;
			}
		}
	}
	return consumed;

}

TBool TInputManager::HasConsumedKeyDown(TInt aScanCode)
{
	TBool hasConsumedKeyDown = EFalse;
	for (TInt i=0 ; i<KMaxKeyRepeats ; i++)
	{
		if (iRepeats[i].iScanCode == aScanCode)
		{
			iRepeats[i].iScanCode = 0;
			hasConsumedKeyDown = ETrue;
			if (iRepeats[i].iRepeatOnFrame)
				StopFrameTicker();
			break;
		}
	}
	return hasConsumedKeyDown;
}


void TInputManager::MarkKeyDown(TInt aScanCode, TBool aWantRepeat)
{
	for (TInt i=0 ; i<KMaxKeyRepeats ; i++)
	{
		if (iRepeats[i].iScanCode == 0)
		{
			iRepeats[i].iScanCode = aScanCode;
			iRepeats[i].iRepeatOnFrame = aWantRepeat ? iFrame + (iFrameTicker->Fps()>>1) : 0;
			if (aWantRepeat)
				StartFrameTicker();
			break;
		}
	}
}


void TInputManager::FrameTick()
{
	iFrame++;
	for (TInt i=0 ; i<KMaxKeyRepeats ; i++)
	{
		if (iRepeats[i].iScanCode && (iRepeats[i].iRepeatOnFrame==iFrame))
		{
			TBool ignore;
			iController->HandleKeyEvent(iRepeats[i].iScanCode, ignore);
			iRepeats[i].iRepeatOnFrame++;
		}
	}
}


void TInputManager::StartFrameTicker()
{
	if (iCount++ == 0)
		iFrameTicker->Start(this);
}

void TInputManager::StopFrameTicker()
{
	if (--iCount == 0)
		iFrameTicker->Stop(this);
}
