/*******************************************************************
 *
 *	File:		ZoomBase.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32base.h>
#include <e32std.h>
#include <e32math.h>
#include <bitdev.h>
#include <GraphicsAccelerator.h>

#include "ZoomBase.h"

#ifdef __WINS__
_LIT(KBitmapFile, "z:\\system\\libs\\zoomanim.mbm");
#else
_LIT(KBitmapFile, "c:\\system\\apps\\ezoom\\zoomanim.mbm");
#endif

/*****************************************************
 *
 * helpers
 *
 ****************************************************/

TUint16* PixelAddress(TUint16* aBitmap, const TPoint& aPoint, TInt aWidth)
{
		// valid assumption for framebuffers & bitmaps
	ASSERT(aPoint.iX>=0 && aPoint.iX<208);
	ASSERT(aPoint.iY>=0 && aPoint.iY<320);

	return aBitmap + (aPoint.iY * aWidth) + aPoint.iX;
}

const TUint16* PixelAddress(const TUint16* aBitmap, const TPoint& aPoint, TInt aWidth)
{
		// valid assumption for framebuffers & bitmaps
	ASSERT(aPoint.iX>=0 && aPoint.iX<208);
	ASSERT(aPoint.iY>=0 && aPoint.iY<320);

	return aBitmap + (aPoint.iY * aWidth) + aPoint.iX;
}

void ConfineRect(TRect& aRect, const TRect& aBoundingRect)
{
	TPoint	delta = aBoundingRect.iTl - aRect.iTl;
	if (delta.iX < 0)
		delta.iX  = 0;
	if (delta.iY < 0)
		delta.iY =  0;
	aRect.Move(delta);
	
	delta = aBoundingRect.iBr - aRect.iBr;
	if (delta.iX > 0)
		delta.iX  = 0;
	if (delta.iY > 0)
		delta.iY =  0;
	aRect.Move(delta);
}

/*
      ****************************
	  *                          *
	  *         1                *
	  *                          *
	  *------******--------------*
	  *      *    *              *
	  *      *    *              *
	  *   2  *  3 *    4         *
	  *      *    *              *
	  *------******--------------*
	  *                          *
	  *                          *
	  *                          *
	  *                          *
	  *                          *
	  *         5                *
	  *                          *
	  *                          *
	  ****************************
*/

void SplitRects(const TRect& aRect, const TRect& aBoundingRect, TRect& aRect1, TRect& aRect2, TRect& aRect3, TRect& aRect4, TRect& aRect5)
{
	aRect1 = TRect(aBoundingRect.iTl.iX, aBoundingRect.iTl.iY, aBoundingRect.iBr.iX, aRect.iTl.iY);
	aRect2 = TRect(aBoundingRect.iTl.iX, aRect.iTl.iY, aRect.iTl.iX, aRect.iBr.iY);
	aRect3 = aRect;
	aRect4 = TRect(aRect.iBr.iX, aRect.iTl.iY, aBoundingRect.iBr.iX, aRect.iBr.iY);
	aRect5 = TRect(aBoundingRect.iTl.iX, aRect.iBr.iY, aBoundingRect.iBr.iX, aBoundingRect.iBr.iY);
}



/*****************************************************
 *
 * class CBitmap
 *
 ****************************************************/


CBitmap::~CBitmap()
{
	User::Free(iBitmap);
}

CBitmap* CBitmap::NewL(const TSize& aSize)
{
	CBitmap*	self = new(ELeave)CBitmap(aSize);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);

	return self;
}

void CBitmap::ConstructL()
{
	TInt bytes = iSize.iWidth * iSize.iHeight * 2;
	iBitmap = (TUint16*) User::AllocL(bytes);
	Mem::FillZ(iBitmap, bytes);
}

void CBitmap::CopyBitmapData(const CFbsBitmap& aFbsbitmap, const TRect& aSrc, const TPoint& aDst)
{
	ASSERT(TRect(aDst, iSize - aDst).Contains(aSrc.iBr - TPoint(1, 1)));

	TUint16*	dstLine = PixelAddress(iBitmap, aDst, iSize.iWidth);

	TInt	endX = aSrc.iBr.iX;
	TInt	endY = aSrc.iBr.iY;
	for (TInt y=aSrc.iTl.iY; y<endY ; y++)
	{
		TUint16*	dst = dstLine;
		for (TInt x=aSrc.iTl.iX ; x<endX ; x++)
		{
			TPoint	p(x, y);
			TRgb	rgbPixel;
			aFbsbitmap.GetPixel(rgbPixel, p);
			*dst = (TUint16) ((aFbsbitmap.DisplayMode() == EColor4K) ? rgbPixel.Color4K() : rgbPixel.Color64K());
			dst++;
		}
		dstLine += iSize.iWidth;
	}
}

/*****************************************************
 *
 * CZoomControl
 *
 ****************************************************/

CZoomControl::CZoomControl(const TSize& aScreenSize, TDisplayMode aDisplayMode)
	: iScreenSize(aScreenSize), iDisplayMode(aDisplayMode)
{
	iVisibleRect = iScreenSize;
}

CZoomControl::~CZoomControl()
{
	delete iBitmap;
	delete iStatusbar;
}


CZoomControl* CZoomControl::NewL(const TSize& aScreenSize, TDisplayMode aDisplayMode)
{
	CZoomControl* self = new(ELeave)CZoomControl(aScreenSize, aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
}

const TInt KToolbarWidth = 104;
const TInt KToolbarHeight = 160 + KButtonHeight;

void CZoomControl::ConstructL()
{
	iBitmap = CBitmap::NewL(TSize(KToolbarWidth, KToolbarHeight));
	iControlRect.SetSize(iBitmap->Size());

	iStatusbar = CTextControl::NewL(TSize(KToolbarWidth, 16), iDisplayMode);
	iStatusbarRect = TRect(0, iControlRect.iBr.iY - iStatusbar->Bitmap()->Size().iHeight, iControlRect.iBr.iX, iControlRect.iBr.iY);

		// toolbar
	iToolbarColor = KRgbWhite;
	for (TInt i =0 ; i< KNoOfButtons ; i++)
		iButtons[i] = TRect(i * KButtonWidth, 0, (i + 1) * KButtonWidth, KButtonHeight);
	iToolBarRect.SetSize(TSize(iBitmap->Size().iWidth, KButtonHeight));
	FillRect(iToolBarRect, iToolbarColor);
	iSlotButtonPressed = -1;

	iScaledUpScreenRect.SetRect(TPoint(0, iToolBarRect.iBr.iY), TSize(iBitmap->Size().iWidth,  iBitmap->Size().iHeight - iToolBarRect.Size().iHeight));
	iScaledDownScreenRect = iScaledUpScreenRect;

	ClearBackground();
}

void CZoomControl::ClearBackground()
{
		// fill background
	FillRect(iScaledUpScreenRect, KRgbRed);
}

void CZoomControl::SetViewPort(const TRect& aRect)
{
	iScaledDownScreenRect = aRect;
	iScaledDownScreenRect.Move(iScaledUpScreenRect.iTl);
	ConfineRect(iScaledDownScreenRect, iScaledUpScreenRect);
}


TBool CZoomControl::IsPointInZoomControl(const TPoint& aPoint) const
{
	return iControlRect.Contains(aPoint);
}

TBool CZoomControl::IsPointInToolBar(const TPoint& aPoint) const
{
	return ToolBarRectOnScreen().Contains(aPoint);
}

TBool CZoomControl::IsPointInScreenRect(const TPoint& aPoint) const
{
	return ScaledDownScreenRectOnScreen().Contains(aPoint);
}

TRect CZoomControl::MoveControl(const TPoint& aDelta)
{
	TRect	rect = iControlRect;
	iControlRect.Move(aDelta);
	ConfineRect(iControlRect, iVisibleRect);
	rect.BoundingRect(iControlRect);
	return rect;
}

void CZoomControl::FillRect(const TRect& aRect, TRgb aColor)
{
	TUint16			pixel = (TUint16) ((iDisplayMode == EColor4K) ? aColor.Color4K() : aColor.Color64K());
	FillRect(aRect, pixel);
}

void CZoomControl::FillRect(const TRect& aRect, TUint16 aColor)
{
	TUint16*	srcLine = PixelAddress(iBitmap->Bitmap(), aRect.iTl, iBitmap->Size().iWidth);
	for(TInt y=0 ; y < aRect.Size().iHeight ; y++)
	{
		TUint16*	src = srcLine;
		for(TInt x=0 ; x < aRect.Size().iWidth ; x++)
			*src++ = aColor;
		srcLine += iBitmap->Size().iWidth;
	}
}


TRect CZoomControl::ToScreenCoord(const TRect& aRect) const
{
	TRect	rect = aRect;
	rect.Move(iControlRect.iTl);
	return rect;
}

void CZoomControl::CopyBitmapData(const CFbsBitmap* aFbsbitmap, CBitmap* aBitmap, TInt aSlot, TBool aPressed)
{
	TPoint	p = iButtons[aSlot].iTl;
	if (aPressed)
		p += TPoint (2, 2);
	TUint16*	dstLine = PixelAddress(aBitmap->Bitmap(), p, aBitmap->Size().iWidth);

	TInt endY = aPressed ? KButtonHeight-2 : KButtonHeight;
	TInt endX = aPressed ? KButtonWidth-2 : KButtonWidth;
	for (TInt y=0; y<endY ; y++)
	{
		TUint16*	dst = dstLine;
		for (TInt x=0 ; x<endX ; x++)
		{
			TPoint	p(x, y);
			TRgb	rgbPixel;
			aFbsbitmap->GetPixel(rgbPixel, p);
			if (rgbPixel != iToolbarColor)
				*dst = (TUint16) ((iDisplayMode == EColor4K) ? rgbPixel.Color4K() : rgbPixel.Color64K());
			dst++;
		}
		dstLine += aBitmap->Size().iWidth;
	}
}

void CZoomControl::LoadButtonBitmap(TInt aSlot, TInt aId, TBool aPressed)
{
	TRAPD(err, LoadButtonBitmapL(aSlot, aId, aPressed));
}

void CZoomControl::LoadButtonBitmapL(TInt aSlot, TInt aId, TBool aPressed)
{
	CFbsBitmap*	bitmap = new(ELeave)CFbsBitmap;
	CleanupStack::PushL(bitmap);

	User::LeaveIfError(bitmap->Load(KBitmapFile, aId));

	FillRect(iButtons[aSlot], iToolbarColor);
	CopyBitmapData(bitmap, iBitmap, aSlot, aPressed);

	CleanupStack::PopAndDestroy(bitmap);
	iButtonsBitmapId[aSlot] = aId;
}

TInt CZoomControl::IsButtonPressed(const TPoint& aPoint) const
{
	TPoint	p = aPoint - iControlRect.iTl;
	TInt slot = -1;
	for (TInt i =0 ; i< KNoOfButtons ; i++)
	{
		if (iButtons[i].Contains(p))
		{
			slot = i;
			break;
		}
	}
	return slot;
}


TBool CZoomControl::SetButtonPressed(TInt aSlot, TBool aPressed)
{
	if ((iSlotButtonPressed == aSlot) && aPressed)
		return EFalse;
	if ((iSlotButtonPressed==-1) && !aPressed)
		return EFalse;

	LoadButtonBitmap(aSlot, iButtonsBitmapId[aSlot], aPressed);

	iSlotButtonPressed = aPressed ? aSlot : -1;
	return ETrue;
}

void CZoomControl::SetBackgroundImage(const CFbsBitmap& aBitmap)
{
	iBitmap->CopyBitmapData(aBitmap, TRect(iScaledUpScreenRect.Size()), iScaledUpScreenRect.iTl);
}


/*****************************************************
 *
 * CTextControl
 *
 ****************************************************/

CTextControl* CTextControl::NewL(const TSize& aSize, TDisplayMode aDisplayMode)
{
	CTextControl* self = new(ELeave)CTextControl(aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL(aSize);
	CleanupStack::Pop(self);
	return self;
}

void CTextControl::SetTextL(const TDesC& aText)
{
	iGc->SetBrushColor(KRgbBlack);
	iGc->SetBrushStyle(CGraphicsContext::ENullBrush);
	iGc->Clear();

	iGc->SetPenColor(KRgbWhite);
	iGc->SetPenSize(TSize(1,1));
	iGc->UseFont(iFont);

	if (aText.Length() > 0)
	{
		TRect		rect(iRawBitmap->Size());
		TInt baselineOffset = (rect.Height() + iFont->AscentInPixels()) / 2; 
		iGc->DrawText(aText, rect, baselineOffset, CGraphicsContext::ECenter);
	}

	iRawBitmap->CopyBitmapData(*iBitmap, TRect(iRawBitmap->Size()), TPoint(0, 0));
}


CTextControl::~CTextControl()
{
	if (iFont)
		iBitmapDevice->ReleaseFont(iFont);
	delete iBitmapDevice;
	delete iBitmap;
	delete iRawBitmap;
}

_LIT(KFontName,"Courier");

void CTextControl::ConstructL(const TSize& aSize)
{
	iBitmap = new(ELeave)CFbsBitmap;
	User::LeaveIfError(iBitmap->Create(aSize, iDisplayMode));
	iBitmapDevice = CFbsBitmapDevice::NewL(iBitmap);
	User::LeaveIfError(iBitmapDevice->CreateContext(iGc));

	TFontSpec fontSpec(KFontName, 20);
	User::LeaveIfError(iBitmapDevice->GetNearestFontInPixels(iFont, fontSpec));
	iGc->UseFont(iFont);

	iRawBitmap = CBitmap::NewL(aSize);
}

