/*******************************************************************
 *
 *	File:		ZoomController.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32base.h>
#include <e32std.h>
#include <e32math.h>
#include <bitdev.h>
#include <GraphicsAccelerator.h>

#include "ZoomController.h"
#include "ZoomDrawer.h"
#include "ZoomBase.h"
#include "FrameTicker.h"
#include "FrameBufferMgr.h"
#include "ReadImage.h"

#include <zoomanim.mbg>


CZoomController::CZoomController(CFrameBufferMgrBase& aFrameBufferMgr, CFrameTicker& aFrameTicker, const CFbsScreenDevice& aFbsScreenDevice)
	: iFrameBufferMgr(aFrameBufferMgr), iFrameTicker(aFrameTicker), iFbsScreenDevice(aFbsScreenDevice), iButtonSlot(-1), iTickContinuously(EFalse)
{
	iModel.Construct(iFbsScreenDevice.SizeInPixels());
}

CZoomController::~CZoomController()
{
	delete iZoomControl;
	iZoomControl = NULL;

	delete iDrawer;
	iDrawer = NULL;
}

CZoomController* CZoomController::NewL(CFrameBufferMgrBase& aFrameBufferMgr, CFrameTicker& aFrameTicker, const CFbsScreenDevice& aFbsScreenDevice)
{
	CZoomController*	self = new(ELeave)CZoomController(aFrameBufferMgr, aFrameTicker, aFbsScreenDevice);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
}

void CZoomController::ConstructL()
{
	iZoomControl = CZoomControl::NewL(iFbsScreenDevice.SizeInPixels(), iFbsScreenDevice.DisplayMode());
	LoadButtonBitmaps();

	iDrawer = new(ELeave)CZoomDrawer(
			(TUint16*) iFrameBufferMgr.FrameBuffer(), 
			(TUint16*) iFrameBufferMgr.ShadowFrameBuffer(),
			iModel,
			*iZoomControl,
			iFbsScreenDevice.DisplayMode()
			);
	IncreaseZoom(50);		// set to 150%
	SetToolActive(ETrue);
}



void CZoomController::FrameTick()
{
	if (iTickContinuously)
		Redraw();

	DrawFrame();

	if (iIsFlashing)
		Flash();
}


void CZoomController::Redraw()
{
	iRedrawRect = TRect(iModel.iScreenSize);
}

void CZoomController::Redraw(const TRect& aRect)
{
	iRedrawRect.BoundingRect(aRect);
	StartTicker();
}

void CZoomController::DrawFrame()
{
//	RDebug::Print(_L("DRAW"));

	if (!iRedrawRect.IsEmpty())
	{
		TRect	rect = iModel.ScaleUp(iRedrawRect);
		rect.iTl.iX -= 1;
		rect.iTl.iY -= 1;
		rect.iBr.iX += 1;
		rect.iBr.iY += 1;
		rect.Intersection(iModel.iViewPort.Size());
		iDrawer->DrawFrame(rect);
		iRedrawRect = TRect(0, 0, 0, 0);
	}
	StopTicker();
}

void CZoomController::MoveViewPort(const TPoint& aDelta)
{
	if (iModel.MoveViewPort(aDelta))
	{
		iZoomControl->SetViewPort(iModel.ScaleViewPort(iZoomControl->ScaledUpScreenRectOnScreen().Size()));
		Redraw(TRect(iModel.iViewPort.Size()));
	}
}

void CZoomController::RedrawScreen()
{
	Redraw(TRect(iModel.iViewPort.Size()));
	iDrawer->ClearBorders();
}

void CZoomController::IncreaseZoom(TInt aDelta)
{
	TInt zoomPercent = iModel.iZoomPercent;

	iModel.IncreaseZoom(aDelta);
	iZoomControl->SetViewPort(iModel.ScaleViewPort(iZoomControl->ScaledUpScreenRectOnScreen().Size()));

	if (zoomPercent != iModel.iZoomPercent)
	{
		TBuf<32>	text;
		text.Format(_L("%i%% zoom"), iModel.iZoomPercent);
		SetStatusbarTextL(text);
	}

	RedrawScreen();
}


void CZoomController::CycleRotation()
{
	iModel.CycleRotation();
	iZoomControl->SetVisibleRect(iModel.VisibleArea());
	iZoomControl->MoveControl(TPoint(0, 0));
	iZoomControl->SetViewPort(iModel.ScaleViewPort(iZoomControl->ScaledUpScreenRectOnScreen().Size()));

	RedrawScreen();
}


TBool CZoomController::HandleKeyEvent(TInt aScanCode, TBool& aWantRepeat)
{
	TBool consumed = EFalse;
	aWantRepeat = EFalse;

	if (iZoomControlActive)
	{	

		switch (aScanCode)
		{
#ifdef __WINS__
			case EStdKeyDelete:
				SetQuality(!iModel.iHighQuality);
				consumed = ETrue;
				break;

			case EStdKeyHome:
				CycleRotation();
				consumed = ETrue;
				aWantRepeat = ETrue;
				break;

			case EStdKeyPageUp:
				IncreaseZoom(25);
				consumed = ETrue;
				aWantRepeat = ETrue;
				break;

			case EStdKeyPageDown:
				IncreaseZoom(-25);
				consumed = ETrue;
				aWantRepeat = ETrue;
				break;
#endif

			case EStdKeyDeviceE:	// jog dial left
				break;

			case EStdKeyDeviceD:	// jog dial right
				break;

			case EStdKeyDevice1:	// jog dial up
				break;

			case EStdKeyDevice2:	// jog dial down
				break;

			case EStdKeyDevice8:	// jog dial select
				break;

			case EStdKeyApplication0:	// browser
				IncreaseZoom(-25);
				consumed = ETrue;
				aWantRepeat = ETrue;
				break;

			case EStdKeyApplication1:	// camera
				IncreaseZoom(25);
				aWantRepeat = ETrue;
				consumed = ETrue;
				break;

			default:
				;
		}
	}

	if (aScanCode == EStdKeyDevice0)	// power
	{
		SetToolActive(!iZoomControlActive);
		consumed = ETrue;
	}

#ifdef __WINS__
	if (aScanCode == EStdKeyEnd)
	{
		SetToolActive(!iZoomControlActive);
		consumed = ETrue;
	}
#endif

	return consumed;
}


TBool CZoomController::HandlePointerDown(const TPoint& aPoint)
{
	TBool	hasConsumed = EFalse;

	if (iZoomControlActive)
	{	
		TPoint p = iModel.RotateBack(aPoint);
		if (iZoomControl->IsPointInZoomControl(p))
		{
			hasConsumed = ETrue;

			iButtonSlot = iZoomControl->IsButtonPressed(p);
			if (iButtonSlot >= 0)
			{
				if (iZoomControl->SetButtonPressed(iButtonSlot, ETrue))
					Redraw(iZoomControl->ButtonRectOnScreen(iButtonSlot));
			}
			else
			{
				iIsDragging = ETrue;
				iIsDraggingViewPort = ((iZoomControl->ScaledUpScreenRectOnScreen().Size() != iZoomControl->ScaledDownScreenRectOnScreen().Size()) && iZoomControl->IsPointInScreenRect(p)) ? ETrue : EFalse;
			}
		}
		iPoint = p;
	}

	return hasConsumed;
}


TBool CZoomController::HandlePointerMove(const TPoint& aPoint)
{
	TBool	hasConsumed = (iIsDragging || (iButtonSlot >= 0));

	if (iZoomControlActive)
	{	
		TPoint p = iModel.RotateBack(aPoint);
		if (iIsDragging)
		{
			if (iIsDraggingViewPort)
			{
				TPoint p1 = p - iPoint;
				p1 = iModel.ScalePointUp(p1, iZoomControl->ScaledUpScreenRectOnScreen().Size());
				MoveViewPort(p1);
			}
			else
			{
				Redraw(iZoomControl->MoveControl(p - iPoint));
			}
			iPoint = p;
		}
		else if (iButtonSlot >= 0)
		{
			if (iZoomControl->SetButtonPressed(iButtonSlot, (iZoomControl->IsButtonPressed(p) == iButtonSlot)))
				Redraw(iZoomControl->ButtonRectOnScreen(iButtonSlot));
		}
	}
	return hasConsumed;
}

TBool CZoomController::HandlePointerUp(const TPoint& aPoint)
{
	TBool	hasConsumed = (iIsDragging || (iButtonSlot >= 0));

	if (iIsDragging)
	{
		HandlePointerMove(aPoint);
		iIsDragging = iIsDraggingViewPort = EFalse;
	}
	else if (iButtonSlot >= 0)
	{
		TPoint p = iModel.RotateBack(aPoint);
		if (iButtonSlot == iZoomControl->IsButtonPressed(p))
		{
			HandleButtonPressed(iButtonSlot);
		}
		if (iZoomControl->SetButtonPressed(iButtonSlot, EFalse))
			Redraw(iZoomControl->ButtonRectOnScreen(iButtonSlot));
		iButtonSlot = -1;
	}

	return hasConsumed;
}

void CZoomController::HandleButtonPressed(TInt aSlotId)
{
	if (aSlotId == ECmdQuality)
	{
		SetQuality(!iModel.iHighQuality);
	}
	else if (aSlotId == ECmdMode)
	{
		SetMode(!iTickContinuously);
	}
	else if (aSlotId == ECmdRotate)
	{
		CycleRotation();
	}
	else if (aSlotId == ECmdClose)
	{
		iCloseRequested = ETrue;
		NotifyApp();
	}
}


TBool CZoomController::ModifyRawEvent(TRawEvent& aRawEvent)
{
	TBool	modified = EFalse;

		// don't modify pointer events when dragging or when toolbar button is pressed
	if (!iIsDragging && (iButtonSlot == -1))
	{
		TRawEvent::TType	type = aRawEvent.Type();
		switch (type)
		{
				// if this pointer event would start a drag operation don't modify it
			case TRawEvent::EButton1Down:
				if (HandlePointerDown(aRawEvent.Pos()))
					break;
				// full through if not
			case TRawEvent::EButton1Up:
			case TRawEvent::EPointerMove:
			{
					/*
					 * We scale the coordinates of the pointer events based 
					 * on the current zoom factor. Note that WServ passes a "const TRawEvent&" 
					 * to the anim. We've comveniently removed the "const" in order 
					 * to modify it.
					 */
				TPoint	point = iModel.ScaleDownAndRotateToViewPort(aRawEvent.Pos());
				TRawEvent& rawEvent = (TRawEvent&) aRawEvent;
				if (point.iX< 0)
					point.iX = -1;
				if (point.iY< 0)
					point.iY = -1;
				rawEvent.Set(type, point.iX, point.iY);
				modified = ETrue;
				break;
			}
			default:
				;
		}
	}
	return modified;
}

void CZoomController::Start()
{
	iRunning = ETrue;
	RedrawScreen();
}

void CZoomController::Stop()
{
	iRunning = EFalse;
	StopTicker();
}

void CZoomController::StartTicker()
{
	if (!iIsTickerStarted)
	{
		iFrameTicker.Start(this);
		iIsTickerStarted = ETrue;
	}
}

TBool CZoomController::KeepingTicking()
{
	return iTickContinuously || iIsFlashing;
}

void CZoomController::StopTicker()
{
	if (iIsTickerStarted)
	{
		if (!(iRunning && KeepingTicking()))
		{
			iFrameTicker.Stop(this);
			iIsTickerStarted = EFalse;
		}
	}
}

void CZoomController::SetToolActive(TBool aActive)
{
	iZoomControlActive = aActive;
	iDrawer->SetZoomControlActive(aActive);
	RedrawScreen();
}



void CZoomController::SetQuality(TBool aHigh)
{
	iModel.SetQuality(aHigh);
	iDrawer->QualityChanged();
	SetStatusbarTextL(aHigh ? _L("High quality") : _L("Low quality"));
	iZoomControl->LoadButtonBitmap(0, aHigh ? EMbmZoomanimHighquality : EMbmZoomanimLowquality);
	RedrawScreen();
}

void CZoomController::SetMode(TBool aAutoMode)
{
	iTickContinuously = aAutoMode;
	SetStatusbarTextL(aAutoMode ? _L("Continuous update") : _L("Smart update"));
	iZoomControl->LoadButtonBitmap(1, aAutoMode  ?  EMbmZoomanimContinuousmode : EMbmZoomanimSmartmode);
	StartTicker();
}

void CZoomController::LoadButtonBitmaps()
{
	iZoomControl->LoadButtonBitmap(0, EMbmZoomanimHighquality);
	iZoomControl->LoadButtonBitmap(1, EMbmZoomanimSmartmode);
	iZoomControl->LoadButtonBitmap(2, EMbmZoomanimCycle);
	iZoomControl->LoadButtonBitmap(3, EMbmZoomanimClose);
}

void CZoomController::NotifyApp()
{
	TRawEvent	event;
	event.Set(TRawEvent::EKeyDown, EStdKeyF15);
	UserSvr::AddEvent(event);

	event.Set(TRawEvent::EKeyUp, EStdKeyF15);
	UserSvr::AddEvent(event);
}

void CZoomController::SetStatusbarTextL(const TDesC& aText)
{
	iText = aText;
	iZoomControl->StatusbarControl()->SetTextL(aText);
	StartFlashing();
}

const TInt KFlashOn = 400;			// 400 ms
const TInt KFlashOff = 200;			// 200 ms
const TInt KFlashPeriod = 2500;		// 2.5 seconds

void CZoomController::Flash()
{
	if (++iFlashFrame == iFlashNextFlashFrame)
	{
		iFlashOn = !iFlashOn;

		if (iFlashOn)
		{
			TRAPD(err, iZoomControl->StatusbarControl()->SetTextL(iText));
		}
		else
		{
			TRAPD(err, iZoomControl->StatusbarControl()->SetTextL(_L("")));
		}
		Redraw(iZoomControl->StatusBarRectOnScreen());

		iFlashFrame = 0;
		iFlashNextFlashFrame = (iFrameTicker.Fps() * (iFlashOn? KFlashOn : KFlashOff)) / 1000;
		if (++iFlashStep == iNoOfFLashSteps)
			StopFlashing();
	}

}

void CZoomController::StopFlashing()
{
	iIsFlashing = EFalse;
	StopTicker();
	iZoomControl->StatusbarControl()->SetHasText(EFalse);
	Redraw(iZoomControl->StatusBarRectOnScreen());
}

void CZoomController::StartFlashing()
{
	iIsFlashing = ETrue;
	iFlashFrame = 0;
	iFlashStep = 0;
	iFlashOn = ETrue;
	iFlashNextFlashFrame = (iFrameTicker.Fps() * KFlashOn) / 1000;
	iNoOfFLashSteps = ((KFlashPeriod / (KFlashOn + KFlashOff)) * 2) +0;
	iZoomControl->StatusbarControl()->SetHasText(ETrue);
	StartTicker();
}



void CZoomController::ReadImageL(const TDesC& aFileName)
{
	delete iReadImage;
	iReadImage = NULL;
	
	iReadImage = CReadImage::NewL(iZoomControl->ScaledUpScreenRectOnScreen().Size(), ETrue, iFbsScreenDevice.DisplayMode(), *this);
	iReadImage->ReadImage(aFileName);
}


void CZoomController::ReadImageComplete(TInt aError)
{
	if (aError == KErrNone)
	{
		iZoomControl->SetBackgroundImage(iReadImage->Bitmap());
		Redraw(iZoomControl->ScaledUpScreenRectOnScreen());
	}
	delete iReadImage;
	iReadImage = NULL;
}


void CZoomController::SetTransparency(TInt aTransparency)
{
	iDrawer->SetTransparency(aTransparency);
	RedrawScreen();
}

void CZoomController::SetBgImage(const TDesC& aFileName)
{
	iZoomControl->ClearBackground();
	TRAPD(err, ReadImageL(aFileName));
	RedrawScreen();
}
