/*******************************************************************
 *
 *	File:		ZoomDrawer.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32base.h>
#include <e32std.h>
#include <e32math.h>
#include <bitdev.h>
#include <GraphicsAccelerator.h>

#include "ZoomDrawer.h"
#include "ZoomModel.h"


CZoomDrawer::CZoomDrawer(TUint16* aFb, TUint16* aShadowFb, TZoomModel& aScaledFullScreenModel, CZoomControl& aZoomControl, TDisplayMode aDisplayMode)
	: iFb(aFb), iShadowFb(aShadowFb), iModel(aScaledFullScreenModel), iZoomControl(aZoomControl), iDisplayMode(aDisplayMode)
{
	QualityChanged();

	iAlpha = 100;
	iFixedColor = (TUint16) ((iDisplayMode == EColor4K) ? KRgbBlue.Color4K() : KRgbBlue.Color64K());
	iTransparentColor = (TUint16) ((iDisplayMode == EColor4K) ? KRgbWhite.Color4K() : KRgbWhite.Color64K());
	iToolbarColor = (TUint16) ((iDisplayMode == EColor4K) ? KRgbRed.Color4K() : KRgbRed.Color64K());
}

void CZoomDrawer::QualityChanged()
{
	iFrameDrawer = iModel.iHighQuality ? ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_BL_RGB444 : &CZoomDrawer::DrawFrame_BL_RGB565) : ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_NN_RGB444 : &CZoomDrawer::DrawFrame_NN_RGB565);
	iFrameAlphaDrawer = iModel.iHighQuality ? ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_BL_RGB444_A : &CZoomDrawer::DrawFrame_BL_RGB565_A) : ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_NN_RGB444_A : &CZoomDrawer::DrawFrame_NN_RGB565_A);
	iFrameTransparentDrawer = iModel.iHighQuality ? ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_BL_RGB444_T : &CZoomDrawer::DrawFrame_BL_RGB565_T) : ((iDisplayMode == EColor4K) ?  &CZoomDrawer::DrawFrame_NN_RGB444_T : &CZoomDrawer::DrawFrame_NN_RGB565_T);
}

void CZoomDrawer::DrawFrame()
{
	TRect	rect(iModel.iViewPort.Size());
	DrawFrame(rect);
}


void CZoomDrawer::DrawAlphaBlendedZoomControl(const TRect& aRect)
{
	TRect	bmScreenRect = iZoomControl.ScaledDownScreenRectOnScreen();
	bmScreenRect.Intersection(aRect);
	if (bmScreenRect.IsEmpty())
		bmScreenRect = TRect(aRect.iTl, TSize(0, 0));

	TRect	r1, r2, r3, r4, r5;
	SplitRects(bmScreenRect, aRect, r1, r2, r3, r4, r5);

	TRect		bmRect = iZoomControl.ControlRectOnScreen();
	CBitmap*	bitmap = iZoomControl.Bitmap();

	if (!r1.IsEmpty())
		(this->*iFrameAlphaDrawer)(r1, bitmap, r1.iTl - bmRect.iTl, EFalse);

	if (!r2.IsEmpty())
		(this->*iFrameAlphaDrawer)(r2, bitmap, r2.iTl - bmRect.iTl, EFalse);

	if (!r3.IsEmpty())
		(this->*iFrameAlphaDrawer)(r3, bitmap, r3.iTl - bmRect.iTl, ETrue);

	if (!r4.IsEmpty())
		(this->*iFrameAlphaDrawer)(r4, bitmap, r4.iTl - bmRect.iTl, EFalse);

	if (!r5.IsEmpty())
		(this->*iFrameAlphaDrawer)(r5, bitmap, r5.iTl - bmRect.iTl, EFalse);
}


void CZoomDrawer::DrawTransparentToolbar(const TRect& aRect)
{
	if (!aRect.IsEmpty())
		(this->*iFrameTransparentDrawer)(aRect, iZoomControl.Bitmap(), aRect.iTl - iZoomControl.ControlRectOnScreen().iTl);
}

void CZoomDrawer::BitBlt(const TRect& aDst, const CBitmap& aBitmap, const TPoint& aPoint)
{
	const TUint16*	srcLine = PixelAddress(aBitmap.Bitmap(), aPoint, aBitmap.Size().iWidth);
	TUint16*		dstLine = PixelAddress(iShadowFb, iModel.MoveToViewPort(iModel.Rotate(aDst.iTl)), iModel.iScreenSize.iWidth);

	TInt	endX = aDst.iBr.iX;
	TInt	endY = aDst.iBr.iY;

	for (TInt y=aDst.iTl.iY; y<endY ; y++)
	{
		const TUint16*	src = srcLine;
		TUint16*		dst = dstLine;

		for (TInt x=aDst.iTl.iX ; x<endX ; x++)
		{
			*dst = *src++;
			dst += iModel.iDirectionX;
		}

		srcLine += aBitmap.Size().iWidth;
		dstLine += iModel.iDirectionY;
	}
}


void CZoomDrawer::DrawZoomControl(const TRect& aRect)
{
	TRect	redrawRect = iZoomControl.ToolBarRectOnScreen();
	redrawRect.Intersection(aRect);
		DrawTransparentToolbar(redrawRect);

	TRect scaledUpScreenRectOnScreen = iZoomControl.ScaledUpScreenRectOnScreen();
	if (iZoomControl.StatusbarControl()->HasText())
	{
		redrawRect = iZoomControl.StatusBarRectOnScreen();
		redrawRect.Intersection(aRect);
		BitBlt(redrawRect, *iZoomControl.StatusbarControl()->Bitmap(), redrawRect.iTl - iZoomControl.StatusBarRectOnScreen().iTl);

		scaledUpScreenRectOnScreen.iBr.iY -= iZoomControl.StatusbarControl()->Bitmap()->Size().iHeight;
	}

	scaledUpScreenRectOnScreen.Intersection(aRect);
	DrawAlphaBlendedZoomControl(scaledUpScreenRectOnScreen);
}


void CZoomDrawer::DrawFrame(const TRect& aRect)
{
	if (!iZoomControlActive)
	{
		(this->*iFrameDrawer)(aRect);
	}
	else
	{
		TRect	rect = aRect;
		rect.Intersection(iZoomControl.ControlRectOnScreen());

		if (rect.IsEmpty())
			(this->*iFrameDrawer)(aRect);
		else
		{
			TRect	r1, r2, r3, r4, r5;
			SplitRects(rect, aRect, r1, r2, r3, r4, r5);

			if (!r1.IsEmpty())
				(this->*iFrameDrawer)(r1);

			if (!r2.IsEmpty())
				(this->*iFrameDrawer)(r2);

			if (!r3.IsEmpty())
				DrawZoomControl(r3);

			if (!r4.IsEmpty())
				(this->*iFrameDrawer)(r4);

			if (!r5.IsEmpty())
				(this->*iFrameDrawer)(r5);
		}
	}
}



TUint16 CZoomDrawer::InterpolatePixelRgb565(TUint16 aPixel0, TUint16 aPixel1, TInt aScale)
{
		// some simple optimisations first
	if (aPixel0 == aPixel1)
		return aPixel0;
	if (aScale<15)
		return aPixel0;
	if (aScale>85)
		return aPixel1;

		// linear interpolisation of pixel aPixel0 and aPixel1
	TInt scale0 = iModel.iScaleTab[100-aScale];
	TInt scale1 = iModel.iScaleTab[aScale];
	TUint	b = ((((aPixel0 & 0x001f) * scale0) + ((aPixel1 & 0x001f) * scale1)) >> 10) & 0x001f;
	TUint	g = ((((aPixel0 & 0x07e0) * scale0) + ((aPixel1 & 0x07e0) * scale1)) >> 10) & 0x07e0;
	TUint	r = ((((aPixel0 & 0xf800) * scale0) + ((aPixel1 & 0xf800) * scale1)) >> 10) & 0xf800;
	return (TUint16) (r | g | b);
}

TUint16 CZoomDrawer::InterpolatePixelRgb444(TUint16 aPixel0, TUint16 aPixel1, TInt aScale)
{
		// some simple optimisations first
	if (aPixel0 == aPixel1)
		return aPixel0;
	if (aScale<15)
		return aPixel0;
	if (aScale>85)
		return aPixel1;

		// linear interpolisation of pixel aPixel0 and aPixel1
	TInt scale0 = iModel.iScaleTab[100-aScale];
	TInt scale1 = iModel.iScaleTab[aScale];
	TUint	b = ((((aPixel0 & 0x00f) * scale0) + ((aPixel1 & 0x00f) * scale1)) >> 10) & 0x00f;
	TUint	g = ((((aPixel0 & 0x0f0) * scale0) + ((aPixel1 & 0x0f0) * scale1)) >> 10) & 0x0f0;
	TUint	r = ((((aPixel0 & 0xf00) * scale0) + ((aPixel1 & 0xf00) * scale1)) >> 10) & 0xf00;
	return (TUint16) (r | g | b);
}

TUint16 CZoomDrawer::AlphaBlendRgb565(TUint16 aPixel0, TUint16 aPixel1, TUint8 aAlpha)
{
	TUint	t = aPixel1 & 0xf800;
	TUint	m1 = (((((aPixel0 & 0xf800) - t) * aAlpha)>>8) & 0xf800) + t;
	t = aPixel1 & 0x07e0;
	TUint	m2 = (((((aPixel0 & 0x07e0) - t) * aAlpha)>>8) & 0x07e0) + t;
	t = aPixel1 & 0x001f;
	TUint	m3 = (((((aPixel0 & 0x001f) - t) * aAlpha)>>8) & 0x001f) + t;
	return (TUint16) (m1 | m2 | m3);
}

TUint16 CZoomDrawer::AlphaBlendRgb444(TUint16 aPixel0, TUint16 aPixel1, TUint8 aAlpha)
{
	TUint	t = aPixel1 & 0x0f00;
	TUint	m1 = (((((aPixel0 & 0x0f00) - t) * aAlpha)>>8) & 0x0f00) + t;
	t = aPixel1 & 0x00f0;
	TUint	m2 = (((((aPixel0 & 0x00f0) - t) * aAlpha)>>8) & 0x00f0) + t;
	t = aPixel1 & 0x000f;
	TUint	m3 = (((((aPixel0 & 0x000f) - t) * aAlpha)>>8) & 0x000f) + t;
	return (TUint16) (m1 | m2 | m3);
}


void CZoomDrawer::ClearBorders()
{
	TRect rect(0, 0, iModel.iScreenSize.iWidth, iModel.iDstPointHeight + 1);
	ClearRect(rect);
	rect = TRect(0, iModel.iScreenSize.iHeight - iModel.iDstPointHeight - 1, iModel.iScreenSize.iWidth, iModel.iScreenSize.iHeight);
	ClearRect(rect);
}

void CZoomDrawer::ClearRect(const TRect& aRect)
{
	TInt yy = aRect.Size().iHeight;
	if (yy == 0)
		return;

	TUint16* line = PixelAddress(iShadowFb, aRect.iTl, iModel.iScreenSize.iWidth);

	for (TInt y=0 ; y<yy ; y++)
	{
		TInt xx = aRect.Size().iWidth;
		TUint16* dst = line;
		while(xx--)
			*dst++ = 0;
		line += iModel.iScreenSize.iWidth;
	}
}


/*************************************************
 *
 * Drawing routines implementated using 
 * pre-precessor MACROs.
 *
 *************************************************/

/*
 * bi-linear scaled drawing routines (high quality)
 */

#define PIXEL InterpolatePixelRgb565(InterpolatePixelRgb565(*src0, *(src0 + 1), dx), InterpolatePixelRgb565(*src1, *(src1 + 1), dx), dy)
void CZoomDrawer::DrawFrame_BL_RGB565(const TRect& aRect)
#include "drawer.xx"
#undef PIXEL

#define PIXEL InterpolatePixelRgb444(InterpolatePixelRgb444(*src0, *(src0 + 1), dx), InterpolatePixelRgb444(*src1, *(src1 + 1), dx), dy)
void CZoomDrawer::DrawFrame_BL_RGB444(const TRect& aRect)
#include "drawer.xx"
#undef PIXEL

#define PIXEL InterpolatePixelRgb565(InterpolatePixelRgb565(*src0, *(src0 + 1), dx), InterpolatePixelRgb565(*src1, *(src1 + 1), dx), dy)
#define BITMAP
#define TRANSPARENT
#define ALPHA(p0,p1,a) AlphaBlendRgb565(p0, p1, a)
void CZoomDrawer::DrawFrame_BL_RGB565_T(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP
#undef TRANSPARENT
#undef PIXEL

#define PIXEL InterpolatePixelRgb444(InterpolatePixelRgb444(*src0, *(src0 + 1), dx), InterpolatePixelRgb444(*src1, *(src1 + 1), dx), dy)
#define BITMAP
#define TRANSPARENT
#define ALPHA(p0,p1,a) AlphaBlendRgb444(p0, p1, a)
void CZoomDrawer::DrawFrame_BL_RGB444_T(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP
#undef TRANSPARENT
#undef PIXEL


#define PIXEL InterpolatePixelRgb565(InterpolatePixelRgb565(*src0, *(src0 + 1), dx), InterpolatePixelRgb565(*src1, *(src1 + 1), dx), dy)
#define BITMAP
#define ALPHA(p0,p1,a) AlphaBlendRgb565(p0, p1, a)
void CZoomDrawer::DrawFrame_BL_RGB565_A(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint, TBool aExtraAlpha)
#include "drawer.xx"
#undef BITMAP
#undef ALPHA
#undef PIXEL

#define PIXEL InterpolatePixelRgb444(InterpolatePixelRgb444(*src0, *(src0 + 1), dx), InterpolatePixelRgb444(*src1, *(src1 + 1), dx), dy)
#define BITMAP
#define ALPHA(p0,p1,a) AlphaBlendRgb444(p0, p1, a)
void CZoomDrawer::DrawFrame_BL_RGB444_A(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint, TBool aExtraAlpha)
#include "drawer.xx"
#undef BITMAP
#undef ALPHA
#undef PIXEL

/*
 * Next-neighbour scaled drawing routines (low quality)
 */
#define PIXEL (dx < 100) ? ((dy < 100) ? *src0 : *src1) : (dy < 100) ?  *(src0+1) : *(src1+1)

void CZoomDrawer::DrawFrame_NN_RGB565(const TRect& aRect)
#include "drawer.xx"

void CZoomDrawer::DrawFrame_NN_RGB444(const TRect& aRect)
#include "drawer.xx"

#define BITMAP
#define TRANSPARENT
#define ALPHA(p0,p1,a) AlphaBlendRgb565(p0, p1, a)
void CZoomDrawer::DrawFrame_NN_RGB565_T(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP
#undef TRANSPARENT


#define BITMAP
#define TRANSPARENT
#define ALPHA(p0,p1,a) AlphaBlendRgb444(p0, p1, a)
void CZoomDrawer::DrawFrame_NN_RGB444_T(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP
#undef TRANSPARENT

#define BITMAP
#define ALPHA(p0,p1,a) AlphaBlendRgb565(p0, p1, a)
void CZoomDrawer::DrawFrame_NN_RGB565_A(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint, TBool aExtraAlpha)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP

#define BITMAP
#define ALPHA(p0,p1,a) AlphaBlendRgb444(p0, p1, a)
void CZoomDrawer::DrawFrame_NN_RGB444_A(const TRect& aRect, const CBitmap* aBitmap, const TPoint& aPoint, TBool aExtraAlpha)
#include "drawer.xx"
#undef ALPHA
#undef BITMAP

