/*******************************************************************
 *
 *	File:		ZoomModel.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32base.h>
#include <e32std.h>
#include <e32math.h>
#include <bitdev.h>
#include <GraphicsAccelerator.h>

#include "ZoomModel.h"

TZoomModel::TZoomModel() : iHighQuality(ETrue)
{
		// fixed point scaling table
	for (TInt i=0 ; i<=100 ; i++)
		iScaleTab[i] = ((1 << 10) * i)/100; 
}


void TZoomModel::Construct(const TSize& aSize)
{
	iScreenSize = aSize;
	SetZoomFactor(100);
}


void TZoomModel::SetZoomFactor(TInt aPercent)
{
	iZoomPercent = aPercent;
	TReal	d;

	Math::Sqrt(d, TReal(aPercent));
	d = TReal(10) / d;
	d *= 100;
	iDelta = (TInt) d;
	ASSERT(iDelta<=100);

	iScaledScreenSize = TSize((iScreenSize.iWidth * 100) / iDelta, (iScreenSize.iHeight * 100) / iDelta);
	iDstPointHeight = 0;
	switch (iRotation)
	{
		default:
		case ERot0:
			iViewPort.SetSize(iScreenSize);
			iDirectionX = 1;
			iDirectionY = iScreenSize.iWidth;
			break;
		case ERot90:
			if (iScaledScreenSize.iWidth > iScreenSize.iHeight)
			{
				iViewPort.SetSize(TSize(iScreenSize.iHeight, iScreenSize.iWidth));
			}
			else
			{
				iViewPort.SetSize(TSize(iScaledScreenSize.iWidth, iScreenSize.iWidth));
				iDstPointHeight = (iScreenSize.iHeight - iScaledScreenSize.iWidth) >> 1;
			}
			iDirectionX = iScreenSize.iWidth;
			iDirectionY = - 1;
			break;
		case ERot180:
			iViewPort.SetSize(iScreenSize);
			iDirectionX = -1;
			iDirectionY = -1 * iScreenSize.iWidth;
			break;
		case ERot270:
			if (iScaledScreenSize.iWidth > iScreenSize.iHeight)
			{
				iViewPort.SetSize(TSize(iScreenSize.iHeight, iScreenSize.iWidth));
			}
			else
			{
				iViewPort.SetSize(TSize(iScaledScreenSize.iWidth, iScreenSize.iWidth));
				iDstPointHeight = (iScreenSize.iHeight - iScaledScreenSize.iWidth) >> 1;
			}
			iDirectionX = iScreenSize.iWidth * -1;
			iDirectionY = 1;
			break;
	}
	MoveViewPort(TPoint(0, 0));		// make sure we fit in screen
}

TBool TZoomModel::IncreaseZoom(TInt aPercent)
{
	TInt old = iZoomPercent;
	iZoomPercent += aPercent;
	if (iZoomPercent < 100)
		iZoomPercent = 100;
	if (iZoomPercent > 1600)
		iZoomPercent = 1600;
	SetZoomFactor(iZoomPercent);
	return (old != iZoomPercent);
}

TBool TZoomModel::MoveViewPort(const TPoint& aDelta)
{
	TPoint	oldTl = iViewPort.iTl;

	iViewPort.Move(aDelta);
	ConfineRect(iViewPort, TRect(iScaledScreenSize));

	return (oldTl != iViewPort.iTl);
}


void TZoomModel::CycleRotation()
{
	iRotation = (TRotation) (iRotation + 1);
	if (iRotation > ERot270)
		iRotation = ERot0;
	SetZoomFactor(iZoomPercent);
}

void TZoomModel::SetQuality(TBool aHigh)
{
	iHighQuality = aHigh;
}

TPoint TZoomModel::RotateBack(const TPoint& aPoint) const
{
	TPoint p(aPoint);
	if (iRotation == ERot90)
		p.iY -= iDstPointHeight;
	else if (iRotation == ERot270)
		p.iY += iDstPointHeight;
	switch (iRotation)
	{
		default:
		case ERot0:
			return p;
			break;
		case ERot90:
			return TPoint(p.iY, (iScreenSize.iWidth - 1) - p.iX);
			break;
		case ERot180:
			return TPoint((iScreenSize.iWidth - 1) - p.iX, (iScreenSize.iHeight - 1) - p.iY);
			break;
		case ERot270:
			return TPoint((iScreenSize.iHeight - 1) - p.iY, p.iX);
	}
	return TPoint(0, 0);
}


TPoint TZoomModel::ScaleUp(const TPoint& aPoint) const
{
	return TPoint((aPoint.iX * 100) / iDelta, (aPoint.iY * 100) / iDelta);
}

TRect TZoomModel::ScaleUp(const TRect& aRect) const
{
	return TRect(ScaleUp(aRect.iTl), ScaleUp(aRect.iBr));
}

void TZoomModel::ScaleDownPrecise(const TPoint& aSrcPoint, TPoint& aDstPoint, TInt& aDx, TInt& aDy) const
{
	aDstPoint = TPoint((aSrcPoint.iX * iDelta) / 100, (aSrcPoint.iY * iDelta) / 100);
	
	aDx = (aSrcPoint.iX * iDelta) - (aDstPoint.iX * 100);
	ASSERT(aDx>=0 && aDx<=100);
	aDy = (aSrcPoint.iY * iDelta) - (aDstPoint.iY * 100);
	ASSERT(aDy>=0 && aDy<=100);
}

TPoint TZoomModel::ScaleDown(const TPoint& aPoint) const
{
	return TPoint((aPoint.iX * iDelta) / 100, (aPoint.iY * iDelta) / 100);
}

TPoint TZoomModel::ScaleDownAndRotateToViewPort(const TPoint& aPoint) const
{
	TPoint p = RotateBack(aPoint);
	return ScaleDown(p + iViewPort.iTl);
}


TPoint TZoomModel::MoveToViewPort(const TPoint& aSrc) const
{
	switch (iRotation)
	{
		default:
		case ERot0:
		case ERot180:
			return aSrc;
			break;
		case ERot90:
			return aSrc + TPoint(0, iDstPointHeight);
			break;
		case ERot270:
			return aSrc - TPoint(0, iDstPointHeight);
	}
	return TPoint(0, 0);
}

TPoint TZoomModel::Rotate(const TPoint& aSrc) const
{
	switch (iRotation)
	{
		default:
		case ERot0:
			return aSrc;
			break;
		case ERot90:
			return TPoint((iScreenSize.iWidth - 1) - aSrc.iY, aSrc.iX);
			break;
		case ERot180:
			return TPoint((iScreenSize.iWidth - 1) - aSrc.iX, (iScreenSize.iHeight - 1) - aSrc.iY);
			break;
		case ERot270:
			return TPoint(aSrc.iY, (iScreenSize.iHeight - 1) - aSrc.iX);
	}
	return TPoint(0, 0);
}


TRect TZoomModel::ScaleViewPort(const TSize& aScaledScreenSize) const
{
	return TRect(ScalePoint(iViewPort.iTl, aScaledScreenSize), ScalePoint(iViewPort.iBr, aScaledScreenSize));
}

TPoint TZoomModel::ScalePoint(const TPoint& aPoint, const TSize& aScaledScreenSize) const
{
	TReal	dx = TReal(aScaledScreenSize.iWidth) / TReal(iScaledScreenSize.iWidth);
	TReal	dy = TReal(aScaledScreenSize.iHeight) / TReal(iScaledScreenSize.iHeight);
	return TPoint(
				(TInt) (TReal(aPoint.iX) * dx),
				(TInt) (TReal(aPoint.iY) * dy)
				);
}

TPoint TZoomModel::ScalePointUp(const TPoint& aPoint, const TSize& aScaledScreenSize) const
{
	TReal	dx = TReal(aScaledScreenSize.iWidth) / TReal(iScaledScreenSize.iWidth);
	TReal	dy = TReal(aScaledScreenSize.iHeight) / TReal(iScaledScreenSize.iHeight);
	return TPoint(
				(TInt) (TReal(aPoint.iX) / dx),
				(TInt) (TReal(aPoint.iY) / dy)
				);
}

TRect TZoomModel::VisibleArea() const
{
	switch (iRotation)
	{
		default:
		case ERot180:
		case ERot0:
			return TRect(iScreenSize);
			break;
		case ERot90:
		case ERot270:
			return TRect(0, 0, iScreenSize.iHeight - (2 * iDstPointHeight), iScreenSize.iWidth);
			break;
	}
	return TRect(iScreenSize);
}
