/*******************************************************************
 *
 *	File:		ZoomEng.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include "ZoomEng.h"
#include <e32keys.h>
#include <quartzkeys.h>
#include <eikenv.h>

const TInt KOptionFileVersion = 1;

const TInt KNotifyKeyCode = EKeyF15;
const TInt KNotifyScanCode = EStdKeyF15;


TInt SystemActivityGenerator(TAny* /*aPtr*/)
{
	TRawEvent	event;
	event.Set(TRawEvent::EActive);
	UserSvr::AddEvent(event);
	return 0;
}


CZoomEngine::CZoomEngine(RWsSession& aWsSession) : iWsSession(aWsSession)
{
}


CZoomEngine::~CZoomEngine()
{
	if (iIsZooming)
		StopZooming();

	DeregisterForNotifications();

	if (iIsLoaded)
	{
		iZoomAnim.Close();
		if (iAnimDll.WsHandle() > 0)
			iAnimDll.Close();
		if (iWindow.WsHandle() > 0)
			iWindow.Close();
		if (iWindowGroup.WsHandle() > 0)
			iWindowGroup.Close();
	}

	delete iSystemActivityTimer;
}

CZoomEngine* CZoomEngine::NewL(RWsSession& aWsSession)
{
	CZoomEngine*	self = new(ELeave)CZoomEngine(aWsSession);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);

	return self;
}


void CZoomEngine::ConstructL()
{
	CEikonEnv::Static()->FsSession().MkDirAll(_L("c:\\system\\apps\\ezoom\\"));
	DefaultOptions();
	LoadOptions();

	RegisterForNotifications();

	iSystemActivityTimer = CPeriodic::NewL(CActive::EPriorityHigh);
}

void CZoomEngine::RegisterForNotifications()
{
	iNotifyScanCode = CEikonEnv::Static()->RootWin().CaptureKeyUpAndDowns(KNotifyScanCode, 0, 0, 255);
	iNotifyKeyCode = CEikonEnv::Static()->RootWin().CaptureKey(KNotifyKeyCode, 0, 0, 255);
}

void CZoomEngine::DeregisterForNotifications()
{
	if (iNotifyScanCode)
		CEikonEnv::Static()->RootWin().CancelCaptureKeyUpAndDowns(iNotifyScanCode);
	if (iNotifyKeyCode)
		CEikonEnv::Static()->RootWin().CancelCaptureKey(iNotifyKeyCode);
}

void CZoomEngine::LoadAnimL()
{
	if (!iIsLoaded)
	{ 
		iWindowGroup = RWindowGroup(iWsSession);
		User::LeaveIfError(iWindowGroup.Construct((TUint32)&iWindowGroup));

		iWindow = RWindow(iWsSession);
		User::LeaveIfError(iWindow.Construct(iWindowGroup, (TUint32)&iWindow));

		TFindFile	findFile(CEikonEnv::Static()->FsSession());
		User::LeaveIfError(findFile.FindByDir(ZOOM_ANIM_NAME, _L("\\system\\libs\\")));
		iAnimDll = RAnimDll(iWsSession);
		User::LeaveIfError(iAnimDll.Load(findFile.File()));
		iZoomAnim = RZoomAnim(iAnimDll);
		User::LeaveIfError(iZoomAnim.Construct(iWindow));

		iIsLoaded = ETrue;
	}
}


void CZoomEngine::SaveOptions(const TZoomOptions& aZoomOptions)
{
	RFile	file;
	TInt err = file.Replace(CEikonEnv::Static()->FsSession(), _L("c:\\system\\apps\\ezoom\\ezoom.ini"), EFileRead);
	if (err != KErrNone)
		err = file.Create(CEikonEnv::Static()->FsSession(), _L("c:\\system\\apps\\ezoom\\ezoom.ini"), EFileRead);
	if (err == KErrNone)
	{
		TInt	version = KOptionFileVersion;
		TPckg<TInt>	versionPkg(version);
		err = file.Write(versionPkg);
		if (err == KErrNone)
		{
			TPckg<TZoomOptions>	optionsPkg(aZoomOptions);
			err = file.Write(optionsPkg);
			file.Close();
		}
	}

	if (err == KErrNone)
	{
		iZoomOptions = aZoomOptions;
		User::InfoPrint(_L("Settings saved"));
	}
	else
		User::InfoPrint(_L("Could not save settings"));

	if (iIsZooming)
		ConfigureAnim();
}


void CZoomEngine::DefaultOptions()
{
	iZoomOptions.iTransparency = 100;
	iZoomOptions.iBgImageFilename = _L("c:\\Documents\\Media files\\image\\unfiled\\cats.jpg");
}


void CZoomEngine::LoadOptions()
{
	RFile	file;
	TInt err = file.Open(CEikonEnv::Static()->FsSession(), _L("c:\\system\\apps\\ezoom\\ezoom.ini"), EFileRead);
	if (err == KErrNone)
	{
		TInt	version;
		TPckg<TInt>	versionPkg(version);
		err = file.Read(versionPkg);
		if (err == KErrNone)
		{
			if (version == 1)
			{
				TPckg<TZoomOptions>	optionsPkg(iZoomOptions);
				err = file.Read(optionsPkg);
			}
		}
		file.Close();
	}
}


void CZoomEngine::StartStopZooming()
{
	if (iIsZooming)
		StopZooming();
	else
		StartZooming();
}


void CZoomEngine::ConfigureAnim()
{
	TZoomAnimConfig	zoomAnimConfig;

	zoomAnimConfig.iBgImageFileName = iZoomOptions.iBgImageFilename;
	zoomAnimConfig.iTransparency = iZoomOptions.iTransparency ;
	
	iZoomAnim.ConfigureAnim(zoomAnimConfig);
}

void CZoomEngine::StartZoomingL()
{
	CEikonEnv::Static()->SetSystem(ETrue);

	LoadAnimL();

	iSystemActivityTimer->Start(TTimeIntervalMicroSeconds32(0), TTimeIntervalMicroSeconds32(1000*1000*5), TCallBack(SystemActivityGenerator));

	ConfigureAnim();
	iZoomAnim.StartAnim(0);
	iIsZooming = ETrue;
}

void CZoomEngine::StartZooming()
{
	TRAPD(err, StartZoomingL())
	if (err)
	{
		StopZooming();
		TBuf<64> text;
		text.Format(_L("Could not start (err=%d)"), err);
		User::InfoPrint(text);
	}
}


void CZoomEngine::StopZooming()
{
	iSystemActivityTimer->Cancel();

	iZoomAnim.StopAnim();
	iIsZooming = EFalse;
	CEikonEnv::Static()->SetSystem(EFalse);
}


TBool CZoomEngine::ConsumeEvent(const TWsEvent& aEvent)
{
	TBool	consumed = EFalse;
	switch (aEvent.Type())
	{
		case EEventKey:
			if (aEvent.Key()->iCode == KNotifyKeyCode)
				consumed = ETrue;
			break;

		case EEventKeyUp:
		case EEventKeyDown:
//			if (aEvent.Key()->iScanCode == KNotifyScanCode)
//				consumed = ETrue;
			break;

		default:
			break;
	}
	return consumed;
}


#if 0

void CZoomEngine::CaptureKey()
{
//	iZoomingKey = CEikonEnv::Static()->RootWin().ZoomingKey(sHotKeyKeyCodes[iZoomOptions.iKey], 0, 0, 2);
}

void CZoomEngine::CancelCaptureKey()
{
//	if (iZoomingKey)
//		CEikonEnv::Static()->RootWin().CancelZoomingKey(iZoomingKey);
//	iZoomingKey = 0;
}


void CZoomEngine::HandleHotKey(TInt aKeyCode)
{
//	if (iIsZooming && (aKeyCode == sHotKeyKeyCodes[iZoomOptions.iKey]))
//	{
//	}
}

#endif 